package net.ihe.gazelle.transformation.other.authentification;

import net.ihe.gazelle.transformation.other.preferences.ApplicationConfigurationDAO;
import net.ihe.gazelle.transformation.other.preferences.ApplicationConfigurationProvider;
import org.apache.commons.lang.StringUtils;
import org.jasig.cas.client.authentication.AttributePrincipal;
import org.jasig.cas.client.util.AbstractCasFilter;
import org.jasig.cas.client.validation.Assertion;
import org.jboss.seam.ScopeType;
import org.jboss.seam.annotations.In;
import org.jboss.seam.annotations.Install;
import org.jboss.seam.annotations.Name;
import org.jboss.seam.annotations.Scope;
import org.jboss.seam.annotations.intercept.BypassInterceptors;
import org.jboss.seam.contexts.Contexts;
import org.jboss.seam.core.Events;
import org.jboss.seam.security.Identity;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.ejb.Startup;
import javax.faces.context.FacesContext;
import javax.servlet.ServletRequest;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpSession;
import java.security.Principal;
import java.util.Map;
import java.util.StringTokenizer;

import static org.jboss.seam.annotations.Install.APPLICATION;

@Name("org.jboss.seam.security.identity")
@Scope(ScopeType.SESSION)
@Install(precedence = APPLICATION)
@BypassInterceptors
@Startup
public class SSOIdentity extends Identity {

    public static final String AUTHENTICATED_USER = "org.jboss.seam.security.management.authenticatedUser";
    public static final String EVENT_USER_AUTHENTICATED = "org.jboss.seam.security.management.userAuthenticated";
    private static final long serialVersionUID = -631532323964539777L;
    private static final String SILENT_LOGIN = "org.jboss.seam.security.silentLogin";

    private static final Logger LOG = LoggerFactory.getLogger(SSOIdentity.class);

    @In
    private ApplicationConfigurationDAO applicationConfigurationDAO;

    @Override
    public String login() {
        LOG.info("Start login...");
        try {
            this.preAuthenticate();
            if (ApplicationConfigurationProvider.instance().getApplicationWorksWithoutCas()
                    && !ApplicationConfigurationProvider.instance().getIpLogin()) {
                // Any user is granted as admin
                String username = "admin";
                this.acceptExternallyAuthenticatedPrincipal(new IpPrincipal(username));
                this.addRole("admin_role");
                this.getCredentials().setUsername(username);
                this.postAuthenticate();
                return "loggedIn";
            } else if (ApplicationConfigurationProvider.instance().getApplicationWorksWithoutCas()
                    && ApplicationConfigurationProvider.instance().getIpLogin()) {
                Object request = FacesContext.getCurrentInstance().getExternalContext().getRequest();
                if (request instanceof ServletRequest) {
                    ServletRequest servletRequest = (ServletRequest) request;
                    String ipRegexp = ApplicationConfigurationProvider.instance().getIpLoginAdmin();
                    if (ipRegexp == null) {
                        ipRegexp = "";
                    }
                    String remoteAddr = servletRequest.getRemoteAddr();
                    if (remoteAddr.matches(ipRegexp)) {
                        String username = "admin";
                        this.acceptExternallyAuthenticatedPrincipal(new IpPrincipal(username));
                        this.addRole("admin_role");
                        this.getCredentials().setUsername(username);
                        this.postAuthenticate();
                        return "loggedIn";
                    }
                }
            } else if (super.isLoggedIn()) {
                // If authentication has already occurred during this request
                // via a silent login, and login() is explicitly called then we
                // still want to raise the LOGIN_SUCCESSFUL event, and then
                // return.
                if (Contexts.isEventContextActive() && Contexts.getEventContext().isSet(SILENT_LOGIN)) {
                    if (Events.exists()) {
                        Events.instance().raiseEvent(EVENT_LOGIN_SUCCESSFUL);
                    }
                    return "loggedIn";
                }
                if (Events.exists()) {
                    Events.instance().raiseEvent(EVENT_ALREADY_LOGGED_IN);
                }
                return "loggedIn";
            }

            AttributePrincipal attributePrincipal = null;
            Principal casPrincipal;
            if ((casPrincipal = this.getCASPrincipal()) == null) {
                return null;
            }
            if (casPrincipal instanceof AttributePrincipal) {
                attributePrincipal = (AttributePrincipal) casPrincipal;
            }

            if (casPrincipal.getName() != null) {
                this.preAuthenticate();

                String username = casPrincipal.getName();

                LOG.info("Found CAS principal for " + username + ": authenticated");
                Map attributes = attributePrincipal.getAttributes();

                this.acceptExternallyAuthenticatedPrincipal(casPrincipal);
                if ((attributes != null) && (attributes.get("role_name") != null)) {
                    String roles = (String) attributes.get("role_name");
                    StringTokenizer st = new StringTokenizer(roles, "[,]");
                    String role;
                    while (st.hasMoreElements()) {
                        role = (String) st.nextElement();
                        role = StringUtils.trimToNull(role);
                        if (role != null) {
                            this.addRole(role);
                        }
                    }
                }
                this.getCredentials().setUsername(username);
                this.postAuthenticate();
                return "loggedIn";

            }

        } catch (RuntimeException e) {
            this.unAuthenticate();
            throw e;
        }

        return null;
    }

    @Override
    public boolean isLoggedIn() {
        if (!super.isLoggedIn()) {
            if (this.getCASPrincipal() != null) {
                this.login();
            }
        }
        return super.isLoggedIn();
    }

    @Override
    public void logout() {
        Object request = FacesContext.getCurrentInstance().getExternalContext().getRequest();
        if (request instanceof ServletRequest) {
            ServletRequest servletRequest = (ServletRequest) request;
            servletRequest.setAttribute(AbstractCasFilter.CONST_CAS_ASSERTION, null);
        }
        Object session = FacesContext.getCurrentInstance().getExternalContext().getSession(false);
        if (session instanceof HttpSession) {
            HttpSession httpSession = (HttpSession) session;
            httpSession.setAttribute(AbstractCasFilter.CONST_CAS_ASSERTION, null);
        }
        super.logout();
    }

    private Principal getCASPrincipal() {
        final HttpServletRequest request = (HttpServletRequest) FacesContext.getCurrentInstance().getExternalContext()
                .getRequest();
        final HttpSession session = request.getSession(false);
        final Assertion assertion = (Assertion) (session == null ? request
                .getAttribute(AbstractCasFilter.CONST_CAS_ASSERTION) : session
                .getAttribute(AbstractCasFilter.CONST_CAS_ASSERTION));
        return assertion == null ? null : assertion.getPrincipal();
    }

}